// Bibliotheken zum Einbinden (Zufallszahlen, Ein-Ausgabe)
#include <cstdlib>
#include <stdio.h>
#include <iostream>
#include "hilf.h"
#include "vector.h"
#include "CSRmatrix.h"
#include "vector2.h"
#include "matrixvbz.h"
#include "LGSdirekt.h"
#include "densematrix.h"
#include "LGSiterativ.h"
#include "graphik.h"

using namespace std;

// 
void exakt(double a,double c,double* y,int m)
// Exakte Loesung von
// -a u''+c u =1 mit m inneren  Gitterpunkten
// wird in den Gitterpunkten berechnet
// Ausgabe: In Vektor y
{
  int i;
  double x4,x1,x2,x3,x0,h;
  if (c>0)
    {
      x1=sqrt(c/a);
      x3=1+exp(-x1);
      h=1./double(m+1);
      x0=h;
      for (i=0;i<m;i++)
	{
          x2=exp(-x1*x0);
          x4=exp(-x1*(1.-x0)); 
          y[i]=1/c*(1.-(x2+x4)/x3);
	  x0=x0+h;
	}
    } 
  else
    {
      h=1./double(m+1);
      x0=h;
      for (i=0;i<m;i++)
	{
	  y[i]=1./(2.*a)*x0*(1.-x0);
           x0=x0+h;
	}
    }
}

// Das Hauptprogramm
int main()
{ 
  int* column_indices;
  int* row_offsets;
  int* vbzcolumn_indices;
  int* vbzrow_indices;
  int* vbzrow_offsets;
  int i,d,n,nz,nvbz;
  double* entries;
  double* vbzentries;
  double* diag;
  double* b;
  double* x;
  double* A;
  char solver,ende,err;
  int m[3],e[1],it;
  double a[4],eins,null,eps,h;
  // Eingabe des Problems 1D-3D, Koeffzienten und Genauigkeit der Diskretisierung
  cout <<"Dimension"<< endl;
  cin>> d;
  e[0]=1;
  vicopy(3,m,1,e,0);
  for (i=0;i<d;i++)
    {
      cout <<"Gitterpunkte in "<<i+1<<"te Richtung (0,1)"<< endl; 
      cin >> m[i]; 
      cout <<"Koeffizient in "<<i+1<<"te Richtung (0,1)"<< endl; 
      cin >> a[i]; 
    } 
   cout <<"Koeffizient vor Term nullter Ordnung "<< endl; 
   cin >> a[i]; 
   n=m[0]*m[1]*m[2];
   nz=n*(2*d+1);
   // Wahl des direkten Loesers sparse/full
   cout <<"Wahl des Loesers (sparse Gauss/full Gauss/iterativ CG/iterativ Jacobi) (s/f/c/j)"<< endl;
   cin >> solver;
   // Speicher allokieren/ Generierung der matrix in CSR-Format
   column_indices=(int*)malloc(nz*sizeof(int));   
   row_offsets=(int*)malloc((n+1)*sizeof(int));   
   entries=(double*)malloc(nz*sizeof(double));  
   switch (d)
     {
     case 1: gitter1(m[0],column_indices,entries,row_offsets,a);break;
     case 2: gitter2(m[0],m[1],column_indices,entries,row_offsets,a);break;
     case 3: gitter3(m[0],m[1],m[2],column_indices,entries,row_offsets,a);
     } 
   nz=row_offsets[n];
   b=(double*)malloc(n*sizeof(double)); 
   x=(double*)malloc(n*sizeof(double)); 
   eins=1;
   vdcopy(n,b,1,&eins,0);
   //   ausgabe(x,n);
   switch (solver)
     {
     case 's': {
	 // Transformiere Matrix in VBZ-Format und Loese mittels Gauss
         LengthVBZfromCSR(column_indices,row_offsets,n,&nvbz);
         vbzcolumn_indices=(int*)malloc(n*sizeof(int));   
         vbzrow_offsets=(int*)malloc((n+1)*sizeof(int));   
         vbzentries=(double*)malloc(nvbz*sizeof(double)); 
         vbzrow_indices=(int*)malloc(n*sizeof(int)); 
         TransformCSRtoVBZ(column_indices,entries,row_offsets,n,nvbz,nz,
    		     vbzcolumn_indices,vbzentries,vbzrow_offsets,vbzrow_indices);
         GaussVBZ(vbzcolumn_indices,vbzentries,vbzrow_offsets,vbzrow_indices,b,x,n);
         free(vbzentries);
         free(vbzcolumn_indices);
         free(vbzrow_offsets);
         free(vbzrow_indices);   
         break;
       }
     case 'f':
       {
	 // Transformiere Matrix in volles Format und Loese mit Gauss (ohne Pivot)
         A=(double*)malloc((n*n)*sizeof(double)); 
         TransformCSRtoDense(column_indices,entries,row_offsets,n,A);
         GaussDense(A,b,x,n,0);
         free(A);
         break;
       }  
     case 'c':
       {
         null=0;
         vdcopy(n,x,1,&null,0);
         cout <<"rel. Genauigkeit?"<< endl;
         cin >> eps;
         CGDiagCSR(column_indices,entries,row_offsets,x,b,n,1000,eps,&it);
         break;
       }
     case 'j':
       {
         null=0;
         vdcopy(n,x,1,&null,0);
         cout <<"rel. Genauigkeit?"<< endl;
         cin >> eps;
         JacobiCSR(column_indices,entries,row_offsets,x,b,n,1000000,eps,&it);
       }
     }
   // Loesung noch skalieren
   h=1/double(m[0]+1);
   vdmult(n,x,1,&h,0,x,1);
   if (d>1)
     {
       h=1/double(m[1]+1);
       vdmult(n,x,1,&h,0,x,1);
     }
    if (d>2)
     {
       h=1/double(m[2]+1);
       vdmult(n,x,1,&h,0,x,1);
     }
   // evtl. Zeichnung des Fehlers auf b gespeichert 
   err='n';
   if (d==1)
     {
       exakt(a[0],a[1],b,n);  // Berechnung der exakten Loesung
       cout <<"Fehler ?" <<endl;
       cin >>err;
     }
   if (err=='j')
     {
       vdminus(n,b,1,x,1,b,1);
       plotFunction(d,m,n,b); //Darstellung des Fehlers
     }  
   else
     {
      plotFunction(d,m,n,x);
     }
    
   free(x);
   free(b);
   free(entries);
   free(column_indices);
   free(row_offsets);
  return 0;
}
